/**
 * file: circular.c
 *
 * PURPOSE:
 * Do a circular test. See README.txt for description.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <getopt.h>

#include <libpv.h>

# define timersub(a, b, result)						      \
  do {									      \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;			      \
    if ((result)->tv_usec < 0) {					      \
      --(result)->tv_sec;						      \
      (result)->tv_usec += 1000000;					      \
    }									      \
  } while (0)

static int client;
/*
 * maximal client count
 */
static int max_client;

/*
 * if send_with_lock is 1:
 * - generate event
 * - unlock engine's lock
 *
 * if send_with_lock is 0:
 * - unlock engine's lock
 * - generate event
 *
 * See function send_notification()
 */
static int send_with_lock;

/*
 * how many turns until next step
 */
static int loops;
static int32_t *my_var,*follow_var;

/* list of all client's pid */
static pid_t *other_client=NULL;

/* list of all process varialbles */
static int32_t **var_list=NULL;

/**
 * check_params - check given params
 *
 * This program understand some more options. They must given
 * in front of the libpv parameters.
 * -m number of maximum clients. Consider: Must correspond with the XML file!
 * -l (0|1) how to deal with notifications and engine lock
 * -t (0|1|2) what test to run
 */
static int check_params(int argc,char *argv[],pv_engine_attr_t *attr)
{
#if 0
	struct option	long_options[] = {
		{ "maxclients",required_argument, 40, 'm' },
		{ "lock",      required_argument,  0, 'l' },
		{ "test",      required_argument,  0, 't' }
	};

	/* reset getopt */
	optind = 0;

	/* parse command line switches */
	while ((opt = getopt_long(argc, argv, "m:l:t:", long_options, &longindex)) != -1) {
		switch (opt) {
		case 'h':
			show_help(basename(argv[0]));
			return -1;
		case 'v':
			_libpv_debuglevel++;	/* FIXME: move this into pv_engine struct (mkl) */
			break;
		case 'd':
			optdesc = optarg;
			break;
		case 's':
			optstore = optarg;
			break;
		case 'j':
			optjvport = optarg;
			break;
		case '8':
			attr->store_type = PV_STORE_8BIT;
			break;
                default:
                        show_help(argv[0]);
			return -EINVAL;
                }
	}
#endif
	max_client=40;
	send_with_lock=0;


	if (argc < 4) {
		pv_engine_attr_t lattr;
		fprintf(stderr,"\nAt least the libpv arguments should be given! See help below\n");
		pv_engine_attr_init_from_cmdline(&lattr, 1, argv);
		return 1;
	}

	if (attr) {
		if (pv_engine_attr_init_from_cmdline(attr, argc, argv))
			return 1;
	}

	return 0;
}

/**
 * Enter with the lock hold! Its unlocked here!
 */
static void send_notification(pv_engine_t *pv_engine,int32_t *var)
{
	if (send_with_lock) {
		pv_event_generate(pv_engine,var);
		pv_unlock(pv_engine);
	}
	else {
		pv_unlock(pv_engine);
		pv_event_generate(pv_engine,var);
	}
}

/**
 * wait_for - use libpv functionality to wait for a change notification
 * @pv_engine:
 * @my_var: The var to wait for
 *
 * Returns 1 if a notification arrieves for exactly @my_var,
 * 0 elsewhere.
 **/
static int wait_for(pv_engine_t *pv_engine,int32_t *my_var)
{
	void *var;

	pv_event_await_syncronous(pv_engine,&var);

	if (var == my_var)
		return 1;

	return 0;
}

/**
 * register_variables - register variables to do the test
 * @pv_engine: Engine to register from
 * @type: 0=simple circular test, 1=complex circular test
 *
 *
 * var1 ... var(n) are always registered, because
 * they are also used for communication beside
 * the simple circular test
 **/
static int register_variables(pv_engine_t *pv_engine,int type)
{
	char buffer[256];

	if (type == 1) {
		printf("Complex Test not supported yet!\n");
		return 1;
	}
/*
 * simple circular test means:
 * - register two variables from the pool:
 *  1) var[client]
 *  2) var[client+1]
 * Register for notification on
 *  1) var[client]
 */
	snprintf(buffer,sizeof(buffer),"var%d",client);
	if ((my_var=pv_register_int32(pv_engine,buffer)) == NULL) {
		printf("Cannot register %s\n",buffer);
		return 1;
	}
	pv_event_register(pv_engine,my_var);

	if (client == max_client)
		snprintf(buffer,sizeof(buffer),"var%d",1);	/* close the loop */
	else
		snprintf(buffer,sizeof(buffer),"var%d",client+1);

	if ((follow_var=pv_register_int32(pv_engine,buffer)) == NULL) {
		printf("Cannot register %s\n",buffer);
		return 1;
	}

	return 0;
}

/**
 * register_all_variables - register all variables
 *
 * Also set all vars to zero (to check if the test was successfully)
 */
static int register_all_variables(pv_engine_t *pv_engine)
{
	char buffer[256];
	int i;

	var_list=calloc(max_client,sizeof(int32_t));
	if (var_list == NULL)
		return EXIT_FAILURE;

	for (i=0; i < max_client; i++) {
		snprintf(buffer,sizeof(buffer),"var%d",i+1);
		if ((var_list[i]=pv_register_int32(pv_engine,buffer)) == NULL) {
			printf("Cannot register %s\n",buffer);
			return EXIT_FAILURE;
		}
		*var_list[i]=0;
	}

	return 0;
}

/**
 * this is done in client2 ... client<max_client>
 *
 * This is the worker for the simple circular test
 **/
static int do_client_stuff(int argc,char *argv[])
{
	pv_engine_attr_t	attr;
	pv_engine_t			pv_engine;
	int err=EXIT_SUCCESS;

	printf("Client%d starts\n",client);

	if (check_params(argc,argv,&attr))
		return EXIT_FAILURE;
	/*
	 * Connect to the process engine
	 */
	if (pv_init(&pv_engine,&attr))
		return EXIT_FAILURE;

	if (register_variables(&pv_engine,0))
		goto exit_1;

	/*
	 * Only wait for notifications and send
	 * a new notification if one is received
	 */
	do{
		if (wait_for(&pv_engine,my_var))
			/* sleep(1); */
			pv_lock(&pv_engine);
			(*follow_var)++;
			send_notification(&pv_engine,follow_var);
	} while(loops--);

	sleep(1);

exit_1:
	pv_destroy(&pv_engine);
	return err;
}


int main(int argc, char *argv[])
{
	pid_t	pid=0;
	pv_engine_attr_t	attr;
	pv_engine_t			pv_engine;
	struct timeval start,stop,diff;
	int long usApprox=0;
	int i,fail;

	client=1;	/* do it once here */
	loops=10000;

	if (check_params(argc,argv,NULL))
		return EXIT_FAILURE;

	other_client=calloc(max_client,sizeof(pid_t));

	if (other_client == NULL)
		return EXIT_FAILURE;
	/*
	 * fork as many clients as given
	 */
	i=1;
	while (i<max_client) {
		client++;
		pid=fork();

		if (pid == 0)
			break;

		other_client[i++]=pid;
	}

	if (pid == 0)
		do_client_stuff(argc,argv);
	else {
		client=1;
		printf("Client#%d starts\n",client);
		sleep(2);
		if (check_params(argc,argv,&attr))
			return EXIT_FAILURE;
		/*
		* Connect to the process engine
		*/
		if (pv_init(&pv_engine,&attr))
			return EXIT_FAILURE;

		if (register_variables(&pv_engine,0))
			goto exit_1;

		sleep(2);
		/*
		 * register to all variables and
		 * set them to zero
		 */
		if (register_all_variables(&pv_engine))
			goto exit_1;

		printf("Client#%d starts messaging\n",client);
		pv_lock(&pv_engine);
		(*follow_var)++;
		gettimeofday(&start,NULL);
		send_notification(&pv_engine,follow_var);

		do {
			if (wait_for(&pv_engine,my_var))
				gettimeofday(&stop,NULL);
				pv_lock(&pv_engine);
				timersub(&stop,&start,&diff);
				if (usApprox == 0)
					usApprox=(long int)diff.tv_sec*1000000+(long int)diff.tv_usec;
				else {
					usApprox+=(long int)diff.tv_sec*1000000+(long int)diff.tv_usec;
					usApprox/=2;
				}
				if (loops) {
					/*
					 * don't send a notification, when we are
					 * going to leave the loop here!
					 */
					(*follow_var)++;
					gettimeofday(&start,NULL);
					send_notification(&pv_engine,follow_var);
				}
		} while(loops--);

	printf("One turnaround needs approximatly %ld usecs\n",usApprox);
	/*
	 * wait to deliver the last remaining notifications
	 */
	sleep(2);

	fail=0;

	/*
	 * The test passes if all vars have the same value
	 */
	for (i=1; i < max_client; i++) {
		if (*var_list[i-1] != *var_list[i]) {
			printf("-> var%d = %d --- var%d = %d\n",i-1,*var_list[i-1],i,*var_list[i]);
			fail=1;
		}
	}

	printf("#### Test %s\n",fail != 0 ? "failed" : "passed");

	free(other_client);
	free(var_list);

exit_1:
	pv_destroy(&pv_engine);
	}
	return EXIT_SUCCESS;
}

/* end of file circular.c */
